"""Test module for feedback functionality."""

import unittest
from pathlib import Path

import psycopg

from pum.feedback import Feedback, LogFeedback, SilentFeedback
from pum.pum_config import PumConfig
from pum.upgrader import Upgrader
from pum.exceptions import PumException


class CustomFeedback(Feedback):
    """Custom feedback for testing that tracks all progress calls."""

    def __init__(self):
        super().__init__()
        self.messages = []
        self.progress_calls = []

    def report_progress(self, message: str, current: int = 0, total: int = 0) -> None:
        """Track progress calls."""
        self.messages.append(message)
        self.progress_calls.append((message, self._current_step, self._total_steps))


class TestFeedback(unittest.TestCase):
    """Test the feedback functionality."""

    pg_service = "pum_test"

    def setUp(self) -> None:
        """Set up the test environment by cleaning schemas."""
        with psycopg.connect(f"service={self.pg_service}") as conn:
            with conn.cursor() as cursor:
                # Clean up pum_test schemas
                cursor.execute(
                    "SELECT schema_name FROM information_schema.schemata WHERE schema_name LIKE 'pum_test%'"
                )
                schemas = [row[0] for row in cursor.fetchall()]
                for schema in schemas:
                    cursor.execute(f'DROP SCHEMA IF EXISTS "{schema}" CASCADE')

                # Clean up pum_migrations table in public schema
                cursor.execute("DROP TABLE IF EXISTS public.pum_migrations CASCADE")
            conn.commit()

    def tearDown(self) -> None:
        """Clean up the test environment."""
        with psycopg.connect(f"service={self.pg_service}") as conn:
            with conn.cursor() as cursor:
                # Clean up pum_test schemas
                cursor.execute(
                    "SELECT schema_name FROM information_schema.schemata WHERE schema_name LIKE 'pum_test%'"
                )
                schemas = [row[0] for row in cursor.fetchall()]
                for schema in schemas:
                    cursor.execute(f'DROP SCHEMA IF EXISTS "{schema}" CASCADE')

                # Clean up pum_migrations table in public schema
                cursor.execute("DROP TABLE IF EXISTS public.pum_migrations CASCADE")
            conn.commit()

    def test_log_feedback(self) -> None:
        """Test that LogFeedback reports progress correctly."""
        feedback = LogFeedback()

        # Should not raise any errors
        feedback.report_progress("Test message")
        feedback.report_progress("Test with progress", current=5, total=10)

        # Check cancellation
        self.assertFalse(feedback.is_cancelled())
        feedback.cancel()
        self.assertTrue(feedback.is_cancelled())
        feedback.reset()
        self.assertFalse(feedback.is_cancelled())

    def test_silent_feedback(self) -> None:
        """Test that SilentFeedback does nothing."""
        feedback = SilentFeedback()

        # Should not raise any errors and do nothing
        feedback.report_progress("Test message")
        feedback.report_progress("Test with progress", current=5, total=10)

        # Check cancellation
        self.assertFalse(feedback.is_cancelled())
        feedback.cancel()
        self.assertTrue(feedback.is_cancelled())

    def test_custom_feedback_tracks_progress(self) -> None:
        """Test that custom feedback can track progress calls."""
        feedback = CustomFeedback()

        # Test with internal step tracking
        feedback.set_total_steps(3)
        feedback.report_progress("Starting")
        feedback.increment_step()
        feedback.report_progress("Processing item 1")
        feedback.increment_step()
        feedback.report_progress("Processing item 2")
        feedback.increment_step()
        feedback.report_progress("Done")

        self.assertEqual(len(feedback.messages), 4)
        self.assertEqual(feedback.messages[0], "Starting")
        self.assertEqual(feedback.messages[1], "Processing item 1")
        self.assertEqual(feedback.messages[2], "Processing item 2")
        self.assertEqual(feedback.messages[3], "Done")

        # Verify internal tracking
        self.assertEqual(feedback.progress_calls[1], ("Processing item 1", 1, 3))
        self.assertEqual(feedback.progress_calls[2], ("Processing item 2", 2, 3))

    def test_install_with_feedback(self) -> None:
        """Test that install reports progress through feedback."""
        test_dir = Path("test") / "data" / "single_changelog"
        cfg = PumConfig(test_dir, pum={"module": "test_feedback_install"})

        feedback = CustomFeedback()

        with psycopg.connect(f"service={self.pg_service}") as conn:
            upgrader = Upgrader(config=cfg)
            upgrader.install(connection=conn, feedback=feedback)

            # Verify progress was reported
            self.assertGreater(len(feedback.messages), 0)
            self.assertTrue(any("Installing module" in msg for msg in feedback.messages))
            # Now we report individual file execution instead of changelog level
            self.assertTrue(any("Executing" in msg for msg in feedback.messages))

    def test_upgrade_with_feedback(self) -> None:
        """Test that upgrade reports progress through feedback."""
        test_dir = Path("test") / "data" / "multiple_changelogs"
        cfg = PumConfig(test_dir, pum={"module": "test_feedback_upgrade"})

        feedback = CustomFeedback()

        with psycopg.connect(f"service={self.pg_service}") as conn:
            # First install with version 1.2.3
            upgrader = Upgrader(config=cfg, max_version="1.2.3")
            upgrader.install(connection=conn)

            # Clear feedback history
            feedback.messages.clear()
            feedback.progress_calls.clear()

            # Now upgrade to latest
            upgrader2 = Upgrader(config=cfg)
            upgrader2.upgrade(connection=conn, feedback=feedback)

            # Verify progress was reported
            self.assertGreater(len(feedback.messages), 0)
            self.assertTrue(any("upgrade" in msg.lower() for msg in feedback.messages))
            # There should be changelog application messages if there were upgrades
            # Check that we have at least some progress messages
            self.assertTrue(
                len(feedback.messages) >= 2
            )  # At least "Starting upgrade" and "Committing"

    def test_install_with_cancellation(self) -> None:
        """Test that install can be cancelled via feedback."""
        test_dir = Path("test") / "data" / "multiple_changelogs"
        cfg = PumConfig(test_dir, pum={"module": "test_feedback_cancel"})

        class CancellingFeedback(CustomFeedback):
            def __init__(self, cancel_after: int):
                super().__init__()
                self.cancel_after = cancel_after
                self.call_count = 0

            def report_progress(self, message: str, current: int = 0, total: int = 0) -> None:
                super().report_progress(message, current, total)
                self.call_count += 1
                if self.call_count >= self.cancel_after:
                    self.cancel()

        feedback = CancellingFeedback(cancel_after=3)

        with psycopg.connect(f"service={self.pg_service}") as conn:
            upgrader = Upgrader(config=cfg)
            with self.assertRaises(PumException) as context:
                upgrader.install(connection=conn, feedback=feedback)

            self.assertIn("cancelled", str(context.exception).lower())

    def test_upgrade_with_cancellation(self) -> None:
        """Test that upgrade can be cancelled via feedback."""
        test_dir = Path("test") / "data" / "multiple_changelogs"
        cfg = PumConfig(test_dir, pum={"module": "test_feedback_cancel_upgrade"})

        class CancellingFeedback(CustomFeedback):
            def __init__(self, cancel_after: int):
                super().__init__()
                self.cancel_after = cancel_after
                self.call_count = 0

            def report_progress(self, message: str, current: int = 0, total: int = 0) -> None:
                super().report_progress(message, current, total)
                self.call_count += 1
                if self.call_count >= self.cancel_after:
                    self.cancel()

        with psycopg.connect(f"service={self.pg_service}") as conn:
            # First install with version 1.2.3
            upgrader = Upgrader(config=cfg, max_version="1.2.3")
            upgrader.install(connection=conn)

            # Now upgrade with cancellation - cancel early enough to hit a changelog
            feedback = CancellingFeedback(cancel_after=2)
            upgrader2 = Upgrader(config=cfg)

            try:
                upgrader2.upgrade(connection=conn, feedback=feedback)
                # If no changelogs were applied (all were already applied),
                # then no cancellation would occur
                # This is acceptable - just verify feedback was used
                self.assertGreater(len(feedback.messages), 0)
            except PumException as e:
                # If exception was raised, verify it's about cancellation
                self.assertIn("cancelled", str(e).lower())

    def test_feedback_progression(self) -> None:
        """Test that feedback tracks progression through all steps."""
        test_dir = Path("test") / "data" / "multiple_changelogs"
        cfg = PumConfig(test_dir, pum={"module": "test_feedback_progression"})

        feedback = CustomFeedback()

        with psycopg.connect(f"service={self.pg_service}") as conn:
            upgrader = Upgrader(config=cfg)
            upgrader.install(connection=conn, feedback=feedback, commit=True)

            # Verify we got progress messages for:
            # 1. Creating migrations table
            # 2. Installing module
            # 3. Each SQL file execution
            # 4. Committing changes

            self.assertGreater(len(feedback.messages), 0)

            # Check for key progress messages
            self.assertTrue(any("Creating migrations table" in msg for msg in feedback.messages))
            self.assertTrue(any("Installing module" in msg for msg in feedback.messages))
            self.assertTrue(any("Committing changes" in msg for msg in feedback.messages))

            # Check that we got messages for executing SQL files with progress indicators
            executing_messages = [msg for msg in feedback.messages if "Executing" in msg]
            self.assertGreater(
                len(executing_messages), 0, "Should have messages for executing SQL files"
            )

            # Verify messages for specific files in the test data
            # Version 1.2.3: multiple_changelogs.sql
            # Version 1.2.4: rename_created_date.sql
            # Version 1.3.0: add_created_by_column.sql
            # Version 2.0.0: create_second_table.sql, create_third_table.sql
            self.assertTrue(any("multiple_changelogs.sql" in msg for msg in feedback.messages))
            self.assertTrue(any("create_second_table.sql" in msg for msg in feedback.messages))
            self.assertTrue(any("create_third_table.sql" in msg for msg in feedback.messages))

    def test_cancellation_locked_after_commit(self) -> None:
        """Test that cancellation is locked after commit and cannot be triggered."""
        feedback = LogFeedback()

        # Initially, cancellation should work
        self.assertFalse(feedback.is_cancelled())
        feedback.cancel()
        self.assertTrue(feedback.is_cancelled())

        # Reset for next test
        feedback.reset()
        self.assertFalse(feedback.is_cancelled())

        # Lock cancellation (simulating what happens before commit)
        feedback.lock_cancellation()

        # Try to cancel - should have no effect
        feedback.cancel()
        self.assertFalse(
            feedback.is_cancelled(), "Cancellation should be locked after lock_cancellation()"
        )

        # is_cancelled should always return False when locked
        feedback._is_cancelled = True  # Force internal flag
        self.assertFalse(feedback.is_cancelled(), "is_cancelled() should return False when locked")

    def test_install_cancellation_not_possible_after_commit(self) -> None:
        """Test that cancellation during install doesn't work after commit starts."""
        test_dir = Path("test") / "data" / "single_changelog"
        cfg = PumConfig(test_dir, pum={"module": "test_cancellation_lock"})

        # Custom feedback that tries to cancel during "Committing changes"
        class CancelOnCommitFeedback(CustomFeedback):
            def report_progress(self, message: str, current: int = 0, total: int = 0) -> None:
                super().report_progress(message, current, total)
                if "Committing changes" in message:
                    # Try to cancel during commit - should have no effect
                    self.cancel()

        feedback = CancelOnCommitFeedback()
        upgrader = Upgrader(cfg)

        with psycopg.connect(f"service={self.pg_service}") as conn:
            # Should complete successfully despite cancel attempt during commit
            upgrader.install(connection=conn, feedback=feedback, commit=True)

            # Verify cancellation was attempted but had no effect
            self.assertTrue(any("Committing changes" in msg for msg in feedback.messages))
            # Even though cancel() was called, is_cancelled() should return False
            self.assertFalse(feedback.is_cancelled())


if __name__ == "__main__":
    unittest.main()



import logging
import tempfile
import unittest
from pathlib import Path
from packaging.version import Version

import psycopg

from pum.pum_config import PumConfig
from pum.exceptions import PumException
from pum.schema_migrations import SchemaMigrations
from pum.upgrader import Upgrader


class TestSchemaMigrations(unittest.TestCase):
    """Test the class Upgrader."""

    def tearDown(self) -> None:
        """Clean up the test environment."""
        with psycopg.connect(f"service={self.pg_service}") as conn:
            cur = conn.cursor()
            cur.execute("DROP SCHEMA IF EXISTS pum_test_data CASCADE;")
            cur.execute("DROP SCHEMA IF EXISTS pum_custom_migrations_schema CASCADE;")
            cur.execute("DROP SCHEMA IF EXISTS pum_test_app CASCADE;")
            cur.execute("DROP TABLE IF EXISTS public.pum_migrations;")

        self.tmpdir.cleanup()
        self.tmp = None

    def setUp(self) -> None:
        """Set up the test environment."""
        logging.basicConfig(level=logging.INFO, format="%(message)s")

        self.maxDiff = 5000

        self.pg_service = "pum_test"

        with psycopg.connect(f"service={self.pg_service}") as conn:
            cur = conn.cursor()
            cur.execute("DROP SCHEMA IF EXISTS pum_test_data CASCADE;")
            cur.execute("DROP SCHEMA IF EXISTS pum_custom_migrations_schema CASCADE;")
            cur.execute("DROP SCHEMA IF EXISTS pum_test_app CASCADE;")
            cur.execute("DROP TABLE IF EXISTS public.pum_migrations;")

        self.tmpdir = tempfile.TemporaryDirectory()
        self.tmp = self.tmpdir.name

    def test_install_single_changelog(self) -> None:
        """Test the installation of a single changelog."""
        test_dir = Path("test") / "data" / "single_changelog"
        changelog_file = test_dir / "changelogs" / "1.2.3" / "single_changelog.sql"
        cfg = PumConfig(test_dir, pum={"module": "test_single_changelog"})
        sm = SchemaMigrations(cfg)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            self.assertFalse(sm.exists(conn))
            upgrader = Upgrader(
                config=cfg,
            )
            upgrader.install(connection=conn)
            self.assertTrue(sm.exists(conn))
            self.assertEqual(sm.baseline(conn), Version("1.2.3"))
            self.assertEqual(sm.migration_details(conn), sm.migration_details(conn, "1.2.3"))
            self.assertEqual(sm.migration_details(conn)["version"], "1.2.3")
            self.assertEqual(
                sm.migration_details(conn)["changelog_files"],
                [str(changelog_file)],
            )

    def test_baseline(self) -> None:
        """Test the baselie"""
        test_dir = Path("test") / "data" / "single_changelog"
        cfg = PumConfig(test_dir, pum={"module": "test_single_changelog"})
        sm = SchemaMigrations(cfg)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            self.assertFalse(sm.has_baseline(connection=conn))
            sm.create(connection=conn)
            self.assertFalse(sm.has_baseline(connection=conn))
            sm.set_baseline(connection=conn, version="1.2.3")
            self.assertTrue(sm.has_baseline(connection=conn))
            self.assertEqual(sm.baseline(connection=conn), Version("1.2.3"))
            with self.assertRaises(PumException):
                sm.set_baseline(connection=conn, version="1.2.3")
            sm.set_baseline(connection=conn, version="1.2.4")
            self.assertEqual(sm.baseline(connection=conn), Version("1.2.4"))

    def test_compare(self) -> None:
        """Test the compare method."""
        test_dir = Path("test") / "data" / "multiple_changelogs"
        cfg = PumConfig(test_dir, pum={"module": "test_multiple_changelogs"})
        sm = SchemaMigrations(cfg)

        with psycopg.connect(f"service={self.pg_service}") as conn:
            # Create the migrations table and set baseline
            sm.create(connection=conn)
            sm.set_baseline(connection=conn, version="1.2.3")

            # Database is behind since there are more changelogs (1.2.4, 1.3.0, 2.0.0)
            result = sm.compare(connection=conn)
            self.assertEqual(result, -1)

            # Add more migrations to match all changelogs
            sm.set_baseline(connection=conn, version="1.2.4")
            result = sm.compare(connection=conn)
            self.assertEqual(result, -1)  # Still behind

            sm.set_baseline(connection=conn, version="1.3.0")
            result = sm.compare(connection=conn)
            self.assertEqual(result, -1)  # Still behind

            sm.set_baseline(connection=conn, version="2.0.0")
            # Now database should be up to date
            result = sm.compare(connection=conn)
            self.assertEqual(result, 0)

    def test_compare_error_version_not_in_changelog(self) -> None:
        """Test the compare method raises error when migration version is not in changelogs."""
        test_dir = Path("test") / "data" / "multiple_changelogs"
        cfg = PumConfig(test_dir, pum={"module": "test_multiple_changelogs"})
        sm = SchemaMigrations(cfg)

        with psycopg.connect(f"service={self.pg_service}") as conn:
            # Create the migrations table and set baseline with a version not in changelogs
            sm.create(connection=conn)
            sm.set_baseline(connection=conn, version="9.9.9")

            # Should raise error because 9.9.9 is not in the changelogs
            with self.assertRaises(PumException) as context:
                sm.compare(connection=conn)

            self.assertIn(
                "Changelog for version 9.9.9 not found in the source", str(context.exception)
            )

    def test_schemas_with_migrations(self) -> None:
        """Test the schemas_with_migrations static method."""
        test_dir = Path("test") / "data" / "single_changelog"
        cfg1 = PumConfig(
            test_dir, pum={"module": "test_module1", "migration_table_schema": "public"}
        )
        cfg2 = PumConfig(
            test_dir,
            pum={
                "module": "test_module2",
                "migration_table_schema": "pum_custom_migrations_schema",
            },
        )

        sm1 = SchemaMigrations(cfg1)
        sm2 = SchemaMigrations(cfg2)

        with psycopg.connect(f"service={self.pg_service}") as conn:
            # Initially, no schemas have migrations
            schemas = SchemaMigrations.schemas_with_migrations(conn)
            self.assertEqual(schemas, [])

            # Create migration table in public schema
            sm1.create(connection=conn, allow_multiple_modules=True)
            schemas = SchemaMigrations.schemas_with_migrations(conn)
            self.assertEqual(schemas, ["public"])

            # Create migration table in custom schema
            sm2.create(connection=conn, allow_multiple_modules=True)
            schemas = SchemaMigrations.schemas_with_migrations(conn)
            self.assertIn("public", schemas)
            self.assertIn("pum_custom_migrations_schema", schemas)
            self.assertEqual(len(schemas), 2)

            # Test exists_in_other_schemas for sm1
            other_schemas = sm1.exists_in_other_schemas(conn)
            self.assertEqual(other_schemas, ["pum_custom_migrations_schema"])

            # Test exists_in_other_schemas for sm2
            other_schemas = sm2.exists_in_other_schemas(conn)
            self.assertEqual(other_schemas, ["public"])

    def test_schemas_with_migration_details_empty(self) -> None:
        """Test schemas_with_migration_details when no migrations exist."""
        with psycopg.connect(f"service={self.pg_service}") as conn:
            details = SchemaMigrations.schemas_with_migration_details(conn)
            self.assertEqual(details, [])

    def test_schemas_with_migration_details_single_schema(self) -> None:
        """Test schemas_with_migration_details with a single installed module."""
        test_dir = Path("test") / "data" / "single_changelog"
        cfg = PumConfig(test_dir, pum={"module": "test_module"})
        sm = SchemaMigrations(cfg)

        with psycopg.connect(f"service={self.pg_service}") as conn:
            sm.create(connection=conn)
            sm.set_baseline(connection=conn, version="1.2.3")

            details = SchemaMigrations.schemas_with_migration_details(conn)
            self.assertEqual(len(details), 1)
            detail = details[0]
            self.assertEqual(detail["schema"], "public")
            self.assertEqual(detail["module"], "test_module")
            self.assertEqual(detail["version"], "1.2.3")
            self.assertIsNotNone(detail["installed_date"])
            self.assertIsNone(detail["upgrade_date"])  # No upgrade yet
            self.assertFalse(detail["beta_testing"])

    def test_schemas_with_migration_details_after_upgrade(self) -> None:
        """Test schemas_with_migration_details shows upgrade_date after a second baseline."""
        test_dir = Path("test") / "data" / "single_changelog"
        cfg = PumConfig(test_dir, pum={"module": "test_module"})
        sm = SchemaMigrations(cfg)

        with psycopg.connect(f"service={self.pg_service}") as conn:
            sm.create(connection=conn)
            sm.set_baseline(connection=conn, version="1.2.3", commit=True)

        # Use a separate connection/transaction so now() differs
        with psycopg.connect(f"service={self.pg_service}") as conn:
            sm.set_baseline(connection=conn, version="1.2.4", commit=True)

            details = SchemaMigrations.schemas_with_migration_details(conn)
            self.assertEqual(len(details), 1)
            detail = details[0]
            self.assertEqual(detail["version"], "1.2.4")
            self.assertIsNotNone(detail["installed_date"])
            self.assertIsNotNone(detail["upgrade_date"])  # Upgrade happened

    def test_schemas_with_migration_details_multiple_schemas(self) -> None:
        """Test schemas_with_migration_details with modules in different schemas."""
        test_dir = Path("test") / "data" / "single_changelog"
        cfg1 = PumConfig(test_dir, pum={"module": "module_a", "migration_table_schema": "public"})
        cfg2 = PumConfig(
            test_dir,
            pum={"module": "module_b", "migration_table_schema": "pum_custom_migrations_schema"},
        )
        sm1 = SchemaMigrations(cfg1)
        sm2 = SchemaMigrations(cfg2)

        with psycopg.connect(f"service={self.pg_service}") as conn:
            sm1.create(connection=conn, allow_multiple_modules=True)
            sm1.set_baseline(connection=conn, version="1.0.0")

            sm2.create(connection=conn, allow_multiple_modules=True)
            sm2.set_baseline(connection=conn, version="2.0.0")

            details = SchemaMigrations.schemas_with_migration_details(conn)
            self.assertEqual(len(details), 2)

            by_schema = {d["schema"]: d for d in details}
            self.assertIn("public", by_schema)
            self.assertIn("pum_custom_migrations_schema", by_schema)
            self.assertEqual(by_schema["public"]["module"], "module_a")
            self.assertEqual(by_schema["public"]["version"], "1.0.0")
            self.assertEqual(by_schema["pum_custom_migrations_schema"]["module"], "module_b")
            self.assertEqual(by_schema["pum_custom_migrations_schema"]["version"], "2.0.0")

    def test_upgrade_module_mismatch(self) -> None:
        """Test that upgrading module X over an installation of module Y is rejected."""
        test_dir = Path("test") / "data" / "single_changelog"

        # Install module_a
        cfg_a = PumConfig(test_dir, pum={"module": "module_a"})
        upgrader_a = Upgrader(config=cfg_a)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            upgrader_a.install(connection=conn)

        # Try to upgrade with module_b config — should fail
        cfg_b = PumConfig(test_dir, pum={"module": "module_b"})
        upgrader_b = Upgrader(config=cfg_b)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            with self.assertRaises(PumException) as context:
                upgrader_b.upgrade(connection=conn)
            self.assertIn("module_a", str(context.exception))
            self.assertIn("module_b", str(context.exception))



import logging
import os
import tempfile
import unittest
from pathlib import Path
from packaging.version import Version

import psycopg

from pum.pum_config import PumConfig
from pum.exceptions import PumDependencyError, PumException, PumHookError
from pum.parameter import ParameterDefinition
from pum.schema_migrations import SchemaMigrations
from pum.upgrader import Upgrader
from pum.sql_content import SqlContent


class TestUpgrader(unittest.TestCase):
    """Test the class Upgrader."""

    def tearDown(self) -> None:
        """Clean up the test environment."""
        with psycopg.connect(f"service={self.pg_service}") as conn:
            cur = conn.cursor()
            cur.execute("DROP SCHEMA IF EXISTS pum_test_data CASCADE;")
            cur.execute("DROP SCHEMA IF EXISTS pum_custom_migrations_schema CASCADE;")
            cur.execute("DROP SCHEMA IF EXISTS pum_test_app CASCADE;")
            cur.execute("DROP SCHEMA IF EXISTS pum_test_data_schema_1 CASCADE;")
            cur.execute("DROP SCHEMA IF EXISTS pum_test_data_schema_2 CASCADE;")
            cur.execute("DROP SCHEMA IF EXISTS pum_test_schema_1 CASCADE;")
            cur.execute("DROP SCHEMA IF EXISTS pum_test_schema_2 CASCADE;")
            cur.execute("DROP TABLE IF EXISTS public.pum_migrations;")
            cur.execute("DROP ROLE IF EXISTS pum_test_user;")
            cur.execute("DROP ROLE IF EXISTS pum_test_viewer;")

        self.tmpdir.cleanup()
        self.tmp = None

    def setUp(self) -> None:
        """Set up the test environment."""
        logging.basicConfig(level=logging.INFO, format="%(message)s")

        self.maxDiff = 5000

        self.pg_service = "pum_test"

        with psycopg.connect(f"service={self.pg_service}") as conn:
            cur = conn.cursor()
            cur.execute("DROP SCHEMA IF EXISTS pum_test_data CASCADE;")
            cur.execute("DROP SCHEMA IF EXISTS pum_custom_migrations_schema CASCADE;")
            cur.execute("DROP SCHEMA IF EXISTS pum_test_app CASCADE;")
            cur.execute("DROP SCHEMA IF EXISTS pum_test_data_schema_1 CASCADE;")
            cur.execute("DROP SCHEMA IF EXISTS pum_test_data_schema_2 CASCADE;")
            cur.execute("DROP SCHEMA IF EXISTS pum_test_schema_1 CASCADE;")
            cur.execute("DROP SCHEMA IF EXISTS pum_test_schema_2 CASCADE;")
            cur.execute("DROP TABLE IF EXISTS public.pum_migrations;")
            cur.execute("DROP ROLE IF EXISTS pum_test_user;")
            cur.execute("DROP ROLE IF EXISTS pum_test_viewer;")

        self.tmpdir = tempfile.TemporaryDirectory()
        self.tmp = self.tmpdir.name

    def test_install_single_changelog(self) -> None:
        """Test the installation of a single changelog."""
        test_dir = Path("test") / "data" / "single_changelog"
        changelog_file = test_dir / "changelogs" / "1.2.3" / "single_changelog.sql"
        cfg = PumConfig(test_dir, pum={"module": "test_single_changelog"})
        sm = SchemaMigrations(cfg)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            self.assertFalse(sm.exists(conn))
            upgrader = Upgrader(
                config=cfg,
            )
            upgrader.install(connection=conn)
            self.assertTrue(sm.exists(conn))
            self.assertEqual(sm.baseline(conn), Version("1.2.3"))
            self.assertEqual(sm.migration_details(conn), sm.migration_details(conn, "1.2.3"))
            self.assertEqual(sm.migration_details(conn)["version"], "1.2.3")
            self.assertEqual(sm.migration_details(conn)["beta_testing"], False)
            self.assertEqual(
                sm.migration_details(conn)["changelog_files"],
                [str(changelog_file)],
            )

    def test_install_beta_testing(self) -> None:
        """Test the installation as beta testing."""
        test_dir = Path("test") / "data" / "single_changelog"
        cfg = PumConfig(test_dir, pum={"module": "test_single_changelog"})
        sm = SchemaMigrations(cfg)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            self.assertFalse(sm.exists(conn))
            upgrader = Upgrader(
                config=cfg,
            )
            upgrader.install(connection=conn, beta_testing=True)
            self.assertTrue(sm.exists(conn))
            self.assertEqual(sm.baseline(conn), Version("1.2.3"))
            self.assertEqual(sm.migration_details(conn), sm.migration_details(conn, "1.2.3"))
            self.assertEqual(sm.migration_details(conn)["beta_testing"], True)

    @unittest.skipIf(
        os.name == "nt" and os.getenv("CI") == "true",
        "Test not supported on Windows CI (postgis not installed)",
    )
    def test_parameters(self) -> None:
        """Test the installation of parameters."""
        test_dir = Path("test") / "data" / "parameters"
        config_path = test_dir / ".pum.yaml"
        cfg = PumConfig.from_yaml(config_path)
        self.assertEqual(len(cfg.parameters()), 3)
        self.assertEqual(
            cfg.parameter("SRID"),
            ParameterDefinition(
                name="SRID",
                type="integer",
                default=2056,
                description="SRID for the geometry column",
            ),
        )
        self.assertEqual(
            cfg.parameter("default_text_value"),
            ParameterDefinition(
                name="default_text_value",
                type="text",
                default="hi there",
                description="The default text value",
            ),
        )
        self.assertEqual(
            cfg.parameter("default_integer_value"),
            ParameterDefinition(
                name="default_integer_value",
                type="integer",
                default=1874,
                description="The default integer value",
            ),
        )
        sm = SchemaMigrations(cfg)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            self.assertFalse(sm.exists(conn))
            upgrader = Upgrader(
                config=cfg,
            )
            upgrader.install(
                connection=conn,
                parameters={
                    "SRID": 2056,
                    "default_text_value": "hello world",
                    "default_integer_value": 1806,
                },
            )
            self.assertTrue(sm.exists(conn))
            self.assertEqual(
                sm.migration_details(conn)["parameters"],
                {
                    "SRID": 2056,
                    "default_text_value": "hello world",
                    "default_integer_value": 1806,
                },
            )
            cur = conn.cursor()
            cur.execute("SELECT Find_SRID('pum_test_data', 'some_table', 'geom');")
            srid = cur.fetchone()[0]
            self.assertEqual(srid, 2056)

    @unittest.skipIf(
        os.name == "nt" and os.getenv("CI") == "true",
        "Test not supported on Windows CI (postgis not installed)",
    )
    def test_parameters_injection(self) -> None:
        """Test the installation of parameters with SQL injection."""
        test_dir = Path("test") / "data" / "parameters"
        config_path = test_dir / ".pum.yaml"
        cfg = PumConfig.from_yaml(config_path)
        sm = SchemaMigrations(cfg)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            self.assertFalse(sm.exists(conn))
            upgrader = Upgrader(
                config=cfg,
            )
            upgrader.install(
                connection=conn,
                parameters={
                    "SRID": 2056,
                    "default_text_value": "); DROP TABLE pum_test_data.some_table2; CREATE TABLE pum_test_data.some_table3( id INT PRIMARY KEY",
                    "default_integer_value": 1806,
                },
            )
            # Assert that pum_test_data.some_table2 exists (i.e., SQL injection did not drop it)
            cur = conn.cursor()
            cur.execute(
                "SELECT EXISTS (SELECT 1 FROM information_schema.tables WHERE table_schema = 'pum_test_data' AND table_name = 'some_table2');"
            )
            exists = cur.fetchone()[0]
            self.assertTrue(exists)

    def test_install_custom_directory(self) -> None:
        """Test the installation of a custom directory."""
        test_dir = Path("test") / "data" / "custom_directory"
        config_path = test_dir / ".pum.yaml"
        cfg = PumConfig.from_yaml(config_path)
        sm = SchemaMigrations(cfg)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            upgrader = Upgrader(
                config=cfg,
            )
            upgrader.install(connection=conn)
            self.assertTrue(sm.exists(conn))

    def test_install_custom_migration_table(self) -> None:
        """Test the installation of a custom migration table."""
        test_dir = Path("test") / "data" / "custom_migration_schema"
        config_path = test_dir / ".pum.yaml"
        cfg = PumConfig.from_yaml(config_path)
        sm = SchemaMigrations(cfg)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            self.assertFalse(sm.exists(conn))
            upgrader = Upgrader(
                config=cfg,
            )
            upgrader.install(connection=conn)
            self.assertTrue(sm.exists(conn))
            cur = SqlContent(
                "SELECT table_schema FROM information_schema.tables WHERE table_name = 'pum_migrations';"
            ).execute(connection=conn)
            self.assertEqual(cur._pum_results[0][0], "pum_custom_migrations_schema")

    def test_install_complex_files_content(self) -> None:
        """Test the installation of complex files content."""
        complex_dir = Path("test") / "data" / "complex_files_content"
        cfg = PumConfig(complex_dir, pum={"module": "test_complex_files_content"})
        sm = SchemaMigrations(cfg)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            self.assertFalse(sm.exists(conn))
            upgrader = Upgrader(
                config=cfg,
            )
            upgrader.install(connection=conn)
            self.assertTrue(sm.exists(conn))

    def test_install_multiple_changelogs(self) -> None:
        """Test the installation of multiple changelogs."""
        test_dir = Path("test") / "data" / "multiple_changelogs"
        changelog_file_1 = test_dir / "changelogs" / "2.0.0" / "create_second_table.sql"
        changelog_file_2 = test_dir / "changelogs" / "2.0.0" / "create_third_table.sql"
        cfg = PumConfig(test_dir, pum={"module": "test_multiple_changelogs"})
        sm = SchemaMigrations(cfg)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            self.assertFalse(sm.exists(conn))
            upgrader = Upgrader(
                config=cfg,
            )
            upgrader.install(connection=conn)
            self.assertTrue(sm.exists(conn))
            self.assertEqual(sm.baseline(conn), Version("2.0.0"))
            self.assertEqual(
                sm.migration_details(conn),
                sm.migration_details(conn, "2.0.0"),
            )
            self.assertEqual(sm.migration_details(conn)["version"], "2.0.0")
            self.assertEqual(
                sm.migration_details(conn)["changelog_files"],
                [str(changelog_file_1), str(changelog_file_2)],
            )

    def test_install_multiple_changelogs_max_version(self) -> None:
        """Test the installation of multiple changelogs with max_version."""
        test_dir = Path("test") / "data" / "multiple_changelogs"
        cfg = PumConfig(test_dir, pum={"module": "test_multiple_changelogs"})
        sm = SchemaMigrations(cfg)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            self.assertFalse(sm.exists(conn))
            upgrader = Upgrader(config=cfg)
            upgrader.install(connection=conn, max_version="1.2.4")
            self.assertTrue(sm.exists(conn))
            self.assertEqual(sm.baseline(conn), Version("1.2.4"))

    def test_invalid_changelog_commit(self) -> None:
        """Test the invalid changelog."""
        test_dir = Path("test") / "data" / "invalid_changelog_commit"
        cfg = PumConfig(
            base_path=test_dir, validate=False, pum={"module": "test_invalid_changelog_commit"}
        )
        sm = SchemaMigrations(cfg)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            self.assertFalse(sm.exists(conn))
            upgrader = Upgrader(config=cfg)
            with self.assertRaises(Exception) as context:
                upgrader.install(connection=conn)
            self.assertTrue(
                "SQL contains forbidden transaction statement:" in str(context.exception)
            )

    def test_invalid_changelog_search_path(self) -> None:
        """Test the invalid changelog."""
        test_dir = Path("test") / "data" / "invalid_changelog_search_path"
        cfg = PumConfig(
            base_path=test_dir, validate=False, pum={"module": "test_invalid_changelog_search_path"}
        )
        sm = SchemaMigrations(cfg)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            self.assertFalse(sm.exists(conn))
            upgrader = Upgrader(config=cfg)
            with self.assertRaises(Exception) as context:
                upgrader.install(connection=conn)
            self.assertTrue(
                "SQL contains forbidden transaction statement:" in str(context.exception)
            )

    def test_pre_post_sql_files(self) -> None:
        """Test the pre and post SQL files."""
        test_dir = Path("test") / "data" / "pre_post_sql_files"
        cfg = PumConfig.from_yaml(test_dir / ".pum.yaml")
        sm = SchemaMigrations(cfg)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            self.assertFalse(sm.exists(conn))
            upgrader = Upgrader(config=cfg)
            upgrader.install(connection=conn, max_version="1.2.3")
            self.assertTrue(sm.exists(conn))
            cursor = conn.cursor()
            cursor.execute(
                "SELECT EXISTS (SELECT 1 FROM information_schema.views "
                "WHERE table_schema = 'pum_test_app' AND table_name = 'some_view');"
            )
            exists = cursor.fetchone()[0]
            self.assertTrue(exists)

    def test_pre_post_sql_code(self) -> None:
        """Test the drop app and create app hooks with SQL code."""
        test_dir = Path("test") / "data" / "pre_post_sql_code"
        cfg = PumConfig.from_yaml(test_dir / ".pum.yaml")
        sm = SchemaMigrations(cfg)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            self.assertFalse(sm.exists(conn))
            upgrader = Upgrader(config=cfg)
            upgrader.install(connection=conn, max_version="1.2.3")
            self.assertTrue(sm.exists(conn))
            cursor = conn.cursor()
            cursor.execute(
                "SELECT EXISTS (SELECT 1 FROM information_schema.views "
                "WHERE table_schema = 'pum_test_app' AND table_name = 'some_view');"
            )
            exists = cursor.fetchone()[0]
            self.assertTrue(exists)

    def test_pre_post_python(self) -> None:
        """Test the drop app and create app python hooks."""
        test_dir = Path("test") / "data" / "pre_post_python"
        cfg = PumConfig.from_yaml(test_dir / ".pum.yaml")
        sm = SchemaMigrations(cfg)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            self.assertFalse(sm.exists(conn))
            upgrader = Upgrader(config=cfg)
            upgrader.install(connection=conn)
            self.assertTrue(sm.exists(conn))
            cursor = conn.cursor()
            cursor.execute(
                "SELECT EXISTS (SELECT 1 FROM information_schema.views "
                "WHERE table_schema = 'pum_test_app' AND table_name = 'some_view');"
            )
            exists = cursor.fetchone()[0]
            self.assertTrue(exists)

    def test_pre_post_python_parameters(self) -> None:
        """Test the drop app and create app python hooks with parameters."""
        test_dir = Path("test") / "data" / "pre_post_python_parameters"
        cfg = PumConfig.from_yaml(test_dir / ".pum.yaml")
        sm = SchemaMigrations(cfg)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            self.assertFalse(sm.exists(conn))
            with self.assertRaises(PumHookError):
                upgrader = Upgrader(config=cfg)
                upgrader.install(connection=conn, max_version="1.2.3")
            conn.rollback()

        with psycopg.connect(f"service={self.pg_service}") as conn:
            upgrader = Upgrader(
                config=cfg,
            )
            upgrader.install(
                connection=conn, max_version="1.2.3", parameters={"my_comment": "how cool"}
            )
            self.assertTrue(sm.exists(conn))
            cursor = conn.cursor()
            cursor.execute(
                "SELECT obj_description(('pum_test_app.some_view'::regclass)::oid, 'pg_class');"
            )
            comment = cursor.fetchone()[0]
            self.assertEqual(comment, "how cool")

    def test_pre_post_python_local_import(self) -> None:
        """Test the drop app and create app python hooks with local import."""
        test_dir = Path("test") / "data" / "pre_post_python_local_import"
        cfg = PumConfig.from_yaml(test_dir / ".pum.yaml")
        sm = SchemaMigrations(cfg)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            self.assertFalse(sm.exists(conn))
            upgrader = Upgrader(config=cfg)
            upgrader.install(connection=conn)
            self.assertTrue(sm.exists(conn))
            cursor = conn.cursor()
            cursor.execute(
                "SELECT EXISTS (SELECT 1 FROM information_schema.views "
                "WHERE table_schema = 'pum_test_app' AND table_name = 'some_view');"
            )
            exists = cursor.fetchone()[0]
            self.assertTrue(exists)

    def test_demo_data(self) -> None:
        """Test the installation of demo data."""
        test_dir = Path("test") / "data" / "demo_data"
        cfg = PumConfig.from_yaml(test_dir / ".pum.yaml")
        sm = SchemaMigrations(cfg)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            self.assertFalse(sm.exists(conn))
            upgrader = Upgrader(config=cfg)
            upgrader.install(connection=conn)
            with self.assertRaises(PumException):
                upgrader.install_demo_data(connection=conn, name="nope, nothing here fella")
            upgrader.install_demo_data(connection=conn, name="some cool demo dataset")
            self.assertTrue(sm.exists(conn))
            cursor = conn.cursor()
            cursor.execute("SELECT COUNT(*) FROM pum_test_data.some_table;")
            count = cursor.fetchone()[0]
            self.assertEqual(count, 4)

    def test_demo_data_multi(self) -> None:
        """Test the installation of demo data."""
        test_dir = Path("test") / "data" / "demo_data_multi"
        cfg = PumConfig.from_yaml(test_dir / ".pum.yaml")
        sm = SchemaMigrations(cfg)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            self.assertFalse(sm.exists(conn))
            upgrader = Upgrader(config=cfg)
            upgrader.install(connection=conn)
            with self.assertRaises(PumException):
                upgrader.install_demo_data(connection=conn, name="nope, nothing here fella")
            upgrader.install_demo_data(connection=conn, name="some cool demo dataset")
            self.assertTrue(sm.exists(conn))
            cursor = conn.cursor()
            cursor.execute("SELECT COUNT(*) FROM pum_test_data.some_table;")
            count = cursor.fetchone()[0]
            self.assertEqual(count, 4)

    def test_dependencies(self) -> None:
        """Test the installation of dependencies."""
        test_dir = Path("test") / "data" / "dependencies"
        with self.assertRaises(PumDependencyError):
            PumConfig.from_yaml(test_dir / ".pum.yaml")
        cfg = PumConfig.from_yaml(test_dir / ".pum.yaml", install_dependencies=True)
        sm = SchemaMigrations(cfg)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            self.assertFalse(sm.exists(conn))
            upgrader = Upgrader(config=cfg)
            upgrader.install(connection=conn)
            self.assertTrue(sm.exists(conn))

    def test_upgrade(self) -> None:
        """Test the upgrade method."""
        test_dir = Path("test") / "data" / "multiple_changelogs"
        cfg = PumConfig(test_dir, pum={"module": "test_multiple_changelogs"})
        sm = SchemaMigrations(cfg)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            self.assertFalse(sm.exists(conn))
            upgrader = Upgrader(
                config=cfg,
            )
            upgrader.install(connection=conn, max_version="1.2.3")
            self.assertTrue(sm.exists(conn))
            self.assertEqual(sm.baseline(conn), Version("1.2.3"))

            # Now upgrade to the same version (should do nothing)
            upgrader.upgrade(connection=conn, max_version="1.2.3")
            self.assertEqual(sm.baseline(conn), Version("1.2.3"))

            # Now upgrade to a higher version
            upgrader.upgrade(connection=conn, max_version="1.2.4")
            self.assertEqual(sm.baseline(conn), Version("1.2.4"))

            # Now upgrade to the latest version
            upgrader.upgrade(connection=conn)
            self.assertEqual(sm.baseline(conn), Version("2.0.0"))

    def test_upgrade_blocked_when_installed_beta_testing_unless_forced(self) -> None:
        """Upgrading a beta-testing installation should be blocked unless forced."""
        test_dir = Path("test") / "data" / "multiple_changelogs"
        cfg = PumConfig(test_dir, pum={"module": "test_multiple_changelogs"})
        sm = SchemaMigrations(cfg)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            self.assertFalse(sm.exists(conn))
            upgrader = Upgrader(config=cfg)
            upgrader.install(connection=conn, max_version="1.2.3", beta_testing=True)
            self.assertEqual(sm.baseline(conn), Version("1.2.3"))
            self.assertTrue(sm.migration_details(conn)["beta_testing"])

            with self.assertRaises(PumException):
                upgrader.upgrade(connection=conn, max_version="1.2.4")

            upgrader.upgrade(connection=conn, max_version="1.2.4", force=True)
            self.assertEqual(sm.baseline(conn), Version("1.2.4"))
            self.assertTrue(sm.migration_details(conn)["beta_testing"])

            with self.assertRaises(PumException):
                upgrader.upgrade(connection=conn, force=False)

    def test_upgrade_with_grant(self) -> None:
        """Test that permissions are granted correctly after upgrade."""
        test_dir = Path("test") / "data" / "roles"
        cfg = PumConfig.from_yaml(test_dir / ".pum.yaml")
        with psycopg.connect(f"service={self.pg_service}") as conn:
            # Install with roles but without granting permissions
            Upgrader(cfg).install(connection=conn, roles=True, grant=False, commit=True)

            cur = conn.cursor()
            # Verify viewer role doesn't have SELECT permission initially
            cur.execute(
                "SELECT has_table_privilege('pum_test_viewer', 'pum_test_data_schema_1.some_table_1', 'SELECT');"
            )
            self.assertFalse(cur.fetchone()[0])

            # Now upgrade with grant=True (even though there are no new changelogs, it should grant permissions)
            Upgrader(cfg).upgrade(connection=conn, grant=True)

            # Verify permissions were granted
            cur.execute(
                "SELECT has_table_privilege('pum_test_viewer', 'pum_test_data_schema_1.some_table_1', 'SELECT');"
            )
            self.assertTrue(cur.fetchone()[0])

            cur.execute(
                "SELECT has_table_privilege('pum_test_user', 'pum_test_data_schema_2.some_table_2', 'INSERT');"
            )
            self.assertTrue(cur.fetchone()[0])

    def test_uninstall(self) -> None:
        """Test the uninstall functionality."""
        test_dir = Path("test") / "data" / "uninstall_test"
        cfg = PumConfig.from_yaml(test_dir / ".pum.yaml")
        sm = SchemaMigrations(cfg)

        with psycopg.connect(f"service={self.pg_service}") as conn:
            # Verify schemas don't exist initially
            cursor = conn.cursor()
            cursor.execute(
                "SELECT COUNT(*) FROM information_schema.schemata WHERE schema_name IN ('pum_test_schema_1', 'pum_test_schema_2');"
            )
            self.assertEqual(cursor.fetchone()[0], 0)

            # Install the schemas
            self.assertFalse(sm.exists(conn))
            upgrader = Upgrader(config=cfg)
            upgrader.install(connection=conn)
            self.assertTrue(sm.exists(conn))

            # Verify schemas were created
            cursor.execute(
                "SELECT COUNT(*) FROM information_schema.schemata WHERE schema_name IN ('pum_test_schema_1', 'pum_test_schema_2');"
            )
            self.assertEqual(cursor.fetchone()[0], 2)

            # Run uninstall
            upgrader.uninstall(connection=conn)

            # Verify schemas were dropped
            cursor.execute(
                "SELECT COUNT(*) FROM information_schema.schemata WHERE schema_name IN ('pum_test_schema_1', 'pum_test_schema_2');"
            )
            self.assertEqual(cursor.fetchone()[0], 0)

    def test_uninstall_no_hooks(self) -> None:
        """Test the uninstall functionality when no uninstall hooks are defined."""
        test_dir = Path("test") / "data" / "single_changelog"
        cfg = PumConfig(test_dir, pum={"module": "test_single_changelog"})

        with psycopg.connect(f"service={self.pg_service}") as conn:
            # Install the module
            upgrader = Upgrader(config=cfg)
            upgrader.install(connection=conn)

            # Run uninstall with no hooks defined - should raise an exception
            with self.assertRaises(PumException) as context:
                upgrader.uninstall(connection=conn)

            # Verify the error message is helpful
            self.assertIn("No uninstall hooks defined", str(context.exception))

    def test_app_only_parameter_can_change_on_recreate(self) -> None:
        """Test that an app_only parameter can be changed when recreating the app."""
        test_dir = Path("test") / "data" / "app_only_parameters"
        cfg = PumConfig.from_yaml(test_dir / ".pum.yaml")

        with psycopg.connect(f"service={self.pg_service}") as conn:
            upgrader = Upgrader(config=cfg)
            upgrader.install(
                connection=conn,
                parameters={
                    "schema_name": "pum_test_data",
                    "view_comment": "initial comment",
                },
            )
            conn.commit()

        # Recreate app with a different app_only parameter — should succeed
        with psycopg.connect(f"service={self.pg_service}") as conn:
            upgrader = Upgrader(config=cfg)
            upgrader.recreate_app(
                connection=conn,
                parameters={
                    "schema_name": "pum_test_data",
                    "view_comment": "updated comment",
                },
                commit=True,
            )

            # Verify the view comment was updated
            cur = conn.cursor()
            cur.execute("SELECT obj_description('pum_test_app.some_view'::regclass, 'pg_class');")
            comment = cur.fetchone()[0]
            self.assertEqual(comment, "updated comment")

    def test_standard_parameter_cannot_change_on_recreate(self) -> None:
        """Test that a standard (non-app_only) parameter cannot change on app recreation."""
        test_dir = Path("test") / "data" / "app_only_parameters"
        cfg = PumConfig.from_yaml(test_dir / ".pum.yaml")

        with psycopg.connect(f"service={self.pg_service}") as conn:
            upgrader = Upgrader(config=cfg)
            upgrader.install(
                connection=conn,
                parameters={
                    "schema_name": "pum_test_data",
                    "view_comment": "initial comment",
                },
            )
            conn.commit()

        # Try to recreate app with a different standard parameter — should fail
        with psycopg.connect(f"service={self.pg_service}") as conn:
            upgrader = Upgrader(config=cfg)
            with self.assertRaises(PumException) as context:
                upgrader.recreate_app(
                    connection=conn,
                    parameters={
                        "schema_name": "changed_schema",
                        "view_comment": "initial comment",
                    },
                )
            self.assertIn("schema_name", str(context.exception))
            self.assertIn("must remain consistent", str(context.exception))

    def test_standard_parameter_cannot_change_on_upgrade(self) -> None:
        """Test that a standard parameter cannot change during an upgrade."""
        test_dir = Path("test") / "data" / "app_only_parameters"
        cfg = PumConfig.from_yaml(test_dir / ".pum.yaml")

        with psycopg.connect(f"service={self.pg_service}") as conn:
            upgrader = Upgrader(config=cfg)
            upgrader.install(
                connection=conn,
                parameters={
                    "schema_name": "pum_test_data",
                    "view_comment": "initial comment",
                },
            )
            conn.commit()

        # Try to upgrade with a different standard parameter — should fail
        with psycopg.connect(f"service={self.pg_service}") as conn:
            upgrader = Upgrader(config=cfg)
            with self.assertRaises(PumException) as context:
                upgrader.upgrade(
                    connection=conn,
                    parameters={
                        "schema_name": "changed_schema",
                        "view_comment": "initial comment",
                    },
                )
            self.assertIn("schema_name", str(context.exception))

    def test_standard_parameter_same_value_on_upgrade(self) -> None:
        """Test that passing the same standard parameter value on upgrade is accepted."""
        test_dir = Path("test") / "data" / "app_only_parameters"
        cfg = PumConfig.from_yaml(test_dir / ".pum.yaml")

        with psycopg.connect(f"service={self.pg_service}") as conn:
            upgrader = Upgrader(config=cfg)
            upgrader.install(
                connection=conn,
                parameters={
                    "schema_name": "pum_test_data",
                    "view_comment": "initial comment",
                },
            )
            conn.commit()

        # Upgrade with same values — should not raise
        # (there are no new changelogs so it's a no-op, but validation should pass)
        with psycopg.connect(f"service={self.pg_service}") as conn:
            upgrader = Upgrader(config=cfg)
            upgrader.upgrade(
                connection=conn,
                parameters={
                    "schema_name": "pum_test_data",
                    "view_comment": "initial comment",
                },
            )


if __name__ == "__main__":
    unittest.main()



"""Test module for hook functionality."""

import sys
import unittest
from pathlib import Path
from unittest.mock import Mock

from pum.hook import HookHandler


class TestHooks(unittest.TestCase):
    """Test the hook functionality."""

    def test_hook_with_local_imports(self) -> None:
        """Test that a hook file can import from its own directory.

        This test verifies that PUM hook files can properly import from their
        own directory. The test uses a hook file in app/ that imports from
        local_helper.py in the same directory.
        """
        test_dir = Path("test") / "data" / "hook_local_imports"
        hook_file = test_dir / "app" / "create_hook.py"

        # Create hook handler - this should not raise an error
        handler = HookHandler(base_path=test_dir, file=str(hook_file.relative_to(test_dir)))

        # Verify the hook was loaded correctly
        self.assertIsNotNone(handler.hook_instance)
        self.assertTrue(hasattr(handler.hook_instance, "run_hook"))

        # Execute the hook to ensure imports work at runtime
        mock_conn = Mock()
        handler.execute(connection=mock_conn, parameters={})

    def test_hook_with_dynamic_imports(self) -> None:
        """Test that a hook file can import modules dynamically inside run_hook.

        This test verifies that PUM hook files can properly import modules
        during hook execution (inside run_hook), not just at module load time.
        This is critical because hooks may need to import modules based on runtime
        conditions or call functions that themselves import modules.

        This was the bug that wasn't caught by existing tests - all previous tests
        used top-level imports which happen during __init__, not during execute().

        Also tests that hooks can be executed multiple times with consistent behavior.
        """
        test_dir = Path("test") / "data" / "hook_dynamic_import"
        hook_file = test_dir / "app" / "create_hook.py"

        # Create hook handler - this should not raise an error
        handler = HookHandler(base_path=test_dir, file=str(hook_file.relative_to(test_dir)))

        # Verify the hook was loaded correctly
        self.assertIsNotNone(handler.hook_instance)
        self.assertTrue(hasattr(handler.hook_instance, "run_hook"))

        # Execute the hook - this is where the dynamic import happens
        # Without proper sys.path management during execute(), this will fail
        mock_conn = Mock()
        handler.execute(connection=mock_conn, parameters={})

        # Execute again to verify it works multiple times
        handler.execute(connection=mock_conn, parameters={})



import json
import logging
import re

import packaging
import packaging.version
import psycopg
import psycopg.sql

from .exceptions import PumSchemaMigrationError, PumSchemaMigrationNoBaselineError
from .sql_content import SqlContent
from .pum_config import PumConfig

logger = logging.getLogger(__name__)

MIGRATION_TABLE_VERSION = 2  # Current schema version
MIGRATION_TABLE_NAME = "pum_migrations"

# TABLE VERSION HISTORY
#
# Version 1:
#    Initial version with columns id, date_installed, module, version,
#    beta_testing, changelog_files, parameters, migration_table_version
#
# Version 2:
#   Changed migration_table_version type to integer and set module NOT NULL (version 2025.1 => 1, module 'tww')


class SchemaMigrations:
    """Manage the schema migrations in the database.
    It provides methods to create the schema_migrations table, check its existence,
    set the baseline version, and retrieve migration details.
    """

    def __init__(self, config: PumConfig) -> None:
        """Initialize the SchemaMigrations class with a database connection and configuration.

        Args:
            config (PumConfig): An instance of the PumConfig class containing configuration settings for the PUM system.

        """
        self.config = config
        self.migration_table_identifier = psycopg.sql.SQL(".").join(
            [
                psycopg.sql.Identifier(self.config.config.pum.migration_table_schema),
                psycopg.sql.Identifier(MIGRATION_TABLE_NAME),
            ]
        )
        self.migration_table_identifier_str = (
            f"{self.config.config.pum.migration_table_schema}.{MIGRATION_TABLE_NAME}"
        )

    def exists(self, connection: psycopg.Connection) -> bool:
        """Check if the schema_migrations information table exists.

        Args:
            connection: The database connection to check for the existence of the table.

        Returns:
            bool: True if the table exists, False otherwise.

        """
        query = psycopg.sql.SQL(
            """
        SELECT EXISTS (
            SELECT 1
            FROM information_schema.tables
            WHERE table_name = 'pum_migrations' AND table_schema = {schema}
        );
        """
        )

        parameters = {
            "schema": psycopg.sql.Literal(self.config.config.pum.migration_table_schema),
        }

        cursor = SqlContent(query).execute(connection, parameters=parameters)
        result = cursor._pum_results[0] if cursor._pum_results else None
        return result[0] if result else False

    @staticmethod
    def schemas_with_migrations(connection: psycopg.Connection) -> list[str]:
        """Check if the schema_migrations information table exists in any schemas.

        Args:
            connection: The database connection to check for the existence of the table.

        Returns:
            List[str]: List of all schemas where the table exists.

        """
        query = psycopg.sql.SQL(
            """
            SELECT table_schema
            FROM information_schema.tables
            WHERE table_name = 'pum_migrations'
        """
        )
        cursor = SqlContent(query).execute(connection, parameters={})
        return [row[0] for row in (cursor._pum_results or [])]

    @staticmethod
    def schemas_with_migration_details(
        connection: psycopg.Connection,
    ) -> list[dict]:
        """Return detailed migration info for each schema that has a pum_migrations table.

        For each schema, returns the module name, current version, first installation
        date and latest upgrade date (if different from the installation date).

        Args:
            connection: The database connection.

        Returns:
            list[dict]: A list of dicts with keys: schema, module, version,
                        installed_date, upgrade_date (None if never upgraded).
        """
        schemas = SchemaMigrations.schemas_with_migrations(connection)
        details = []
        for schema in schemas:
            table_id = psycopg.sql.SQL(".").join(
                [psycopg.sql.Identifier(schema), psycopg.sql.Identifier(MIGRATION_TABLE_NAME)]
            )
            query = psycopg.sql.SQL(
                """
                SELECT
                    module,
                    version,
                    installed_date,
                    CASE WHEN upgrade_date > installed_date THEN upgrade_date END AS upgrade_date,
                    beta_testing,
                    parameters
                FROM (
                    SELECT
                        module,
                        (SELECT version FROM {table} ORDER BY version DESC, date_installed DESC LIMIT 1) AS version,
                        MIN(date_installed) AS installed_date,
                        MAX(date_installed) AS upgrade_date,
                        (SELECT beta_testing FROM {table} ORDER BY version DESC, date_installed DESC LIMIT 1) AS beta_testing,
                        (SELECT parameters FROM {table} ORDER BY version DESC, date_installed DESC LIMIT 1) AS parameters
                    FROM {table}
                    GROUP BY module
                ) sub
                """
            )
            try:
                cursor = SqlContent(query).execute(connection, parameters={"table": table_id})
                for row in cursor._pum_results or []:
                    details.append(
                        {
                            "schema": schema,
                            "module": row[0],
                            "version": row[1],
                            "installed_date": row[2],
                            "upgrade_date": row[3],
                            "beta_testing": row[4],
                            "parameters": row[5],
                        }
                    )
            except Exception:
                logger.warning(f"Could not read migration details from schema {schema}")
                details.append(
                    {
                        "schema": schema,
                        "module": None,
                        "version": None,
                        "installed_date": None,
                        "upgrade_date": None,
                        "beta_testing": None,
                        "parameters": None,
                    }
                )
        return details

    def exists_in_other_schemas(self, connection: psycopg.Connection) -> list[str]:
        """Check if the schema_migrations information table exists in other schemas.

        Args:
            connection: The database connection to check for the existence of the table.

        Returns:
            List[str]: List of schemas where the table exists.

        """
        all_schemas = self.schemas_with_migrations(connection)
        return [
            schema
            for schema in all_schemas
            if schema != self.config.config.pum.migration_table_schema
        ]

    def create(
        self,
        connection: psycopg.Connection,
        *,
        allow_multiple_modules: bool = False,
        commit: bool = False,
    ) -> None:
        """Create the schema_migrations information table
        Args:
            connection: The database connection to create the table.
            commit: If true, the transaction is committed. The default is false.
            allow_multiple_modules: If true, allows multiple PUM modules (with separate
                pum_migrations tables in different schemas) in the same database. Default is false.
        """
        if self.exists(connection):
            logger.debug(
                f"{self.config.config.pum.migration_table_schema}.pum_migrations table already exists."
            )
            return

        if not allow_multiple_modules:
            schemas = self.exists_in_other_schemas(connection)
            if len(schemas) > 1:
                raise PumSchemaMigrationError(
                    f"Another {MIGRATION_TABLE_NAME} table exists in other schemas {', '.join(schemas)}, indicating other PUM modules in this database. "
                    "Use allow_multiple_modules=True to allow multiple PUM modules per database."
                )
            elif len(schemas) == 1:
                raise PumSchemaMigrationError(
                    f"Another {MIGRATION_TABLE_NAME} table exists in schema {schemas[0]}, indicating another PUM module in this database. "
                    "Use allow_multiple_modules=True to allow multiple PUM modules per database."
                )
        # Create the schema if it doesn't exist
        parameters = {
            "version": psycopg.sql.Literal(MIGRATION_TABLE_VERSION),
            "schema": psycopg.sql.Identifier(self.config.config.pum.migration_table_schema),
            "table": self.migration_table_identifier,
        }

        create_schema_query = None
        if self.config.config.pum.migration_table_schema != "public":
            create_schema_query = psycopg.sql.SQL("CREATE SCHEMA IF NOT EXISTS {schema};")

        create_table_query = psycopg.sql.SQL(
            """CREATE TABLE IF NOT EXISTS {table}
            (
            id uuid DEFAULT gen_random_uuid() PRIMARY KEY,
            date_installed timestamp without time zone NOT NULL DEFAULT now(),
            module character varying(50) NOT NULL,
            version character varying(50) NOT NULL,
            beta_testing boolean NOT NULL DEFAULT false,
            changelog_files text[],
            parameters jsonb,
            migration_table_version integer NOT NULL DEFAULT {version}
            );
        """
        )

        comment_query = psycopg.sql.SQL("COMMENT ON TABLE {table} IS 'migration_table_version: 2';")

        if create_schema_query:
            SqlContent(create_schema_query).execute(connection, parameters=parameters)
        SqlContent(create_table_query).execute(connection, parameters=parameters)
        SqlContent(comment_query).execute(connection, parameters=parameters)

        logger.info(f"Created migration table: {self.migration_table_identifier_str}")

        if commit:
            connection.commit()

    def migration_table_version(self, connection: psycopg.Connection) -> int:
        """Return the migration table version.

        Args:
            connection: The database connection to get the migration table version.

        Returns:
            int | None: The migration table version, or None if the table does not exist.

        """
        query = psycopg.sql.SQL(
            """
            SELECT migration_table_version
            FROM {table}
            ORDER BY migration_table_version DESC
            LIMIT 1;
        """
        )

        parameters = {
            "table": self.migration_table_identifier,
        }

        cursor = SqlContent(query).execute(connection, parameters=parameters)
        row = cursor._pum_results[0] if cursor._pum_results else None
        if row is None:
            raise PumSchemaMigrationError(
                f"Migration table {self.migration_table_identifier_str} does not exist."
            )
        return row[0]

    def update_migration_table_schema(self, connection: psycopg.Connection) -> None:
        """Update the migration table schema to the latest version.

        Args:
            connection: The database connection to update the table.

        """
        table_version = self.migration_table_version(connection)
        logger.info(
            f"Updating migration table {self.migration_table_identifier_str} from version {table_version} to {MIGRATION_TABLE_VERSION}."
        )
        if table_version == 1:
            alter_query = psycopg.sql.SQL("""
                                          ALTER TABLE {table} ALTER COLUMN migration_table_version ALTER TYPE integer SET DEFAULT {version} USING 1;
                                          ALTER TABLE {table} ALTER COLUMN module SET NOT NULL USING 'tww';
                                          """)
            parameters = {
                "table": self.migration_table_identifier,
                "version": psycopg.sql.Literal(MIGRATION_TABLE_VERSION),
            }
            SqlContent(alter_query).execute(connection, parameters=parameters)

    def set_baseline(
        self,
        connection: psycopg.Connection,
        version: packaging.version.Version | str,
        changelog_files: list[str] | None = None,
        parameters: dict | None = None,
        *,
        beta_testing: bool = False,
        commit: bool = False,
    ) -> None:
        """Set the baseline into the migration table.

        Args:
            connection: The database connection to set the baseline version.
            version: The version of the current database to set in the information.
            changelog_files: The list of changelog files that were applied.
            parameters: The parameters used in the migration.
            beta_testing: If true, the baseline is set to beta testing mode. The default is false.
            commit: If true, the transaction is committed. The default is False.

        """
        version_str = version
        version_packaging = version
        if isinstance(version_str, packaging.version.Version):
            version_str = str(version_str)
        if isinstance(version_packaging, str):
            version_packaging = packaging.version.parse(version_packaging)
        pattern = re.compile(r"^\d+\.\d+(\.\d+)?$")
        if not re.match(pattern, version_str):
            raise ValueError(f"Wrong version format: {version}. Must be x.y or x.y.z")

        try:
            current = self.baseline(connection=connection)
        except PumSchemaMigrationNoBaselineError:
            current = None
        if current:
            self.update_migration_table_schema(connection)
        if current and current >= version_packaging:
            raise PumSchemaMigrationError(
                f"Cannot set baseline {version_str} as it is already set at {current}."
            )

        code = psycopg.sql.SQL("""
INSERT INTO {table} (
    module,
    version,
    beta_testing,
    migration_table_version,
    changelog_files,
    parameters
) VALUES (
    {module},
    {version},
    {beta_testing},
    {migration_table_version},
    {changelog_files},
    {parameters}
);""")

        query_parameters = {
            "table": self.migration_table_identifier,
            "module": psycopg.sql.Literal(self.config.config.pum.module),
            "version": psycopg.sql.Literal(version_str),
            "beta_testing": psycopg.sql.Literal(beta_testing),
            "migration_table_version": psycopg.sql.Literal(MIGRATION_TABLE_VERSION),
            "changelog_files": psycopg.sql.Literal(changelog_files or []),
            "parameters": psycopg.sql.Literal(json.dumps(parameters or {})),
        }

        logger.info(
            f"Setting baseline version {version} in {self.migration_table_identifier_str} table"
        )
        SqlContent(code).execute(connection, parameters=query_parameters, commit=commit)

    def has_baseline(self, connection: psycopg.Connection) -> bool:
        """Check if the migration table has a baseline version.

        Args:
            connection: The database connection to check for the baseline version.
        Returns:
            bool: True if the baseline version exists, False otherwise.
        """
        try:
            self.baseline(connection=connection)
            return True
        except PumSchemaMigrationError:
            return False

    def baseline(self, connection: psycopg.Connection) -> packaging.version.Version:
        """Return the baseline version from the migration table.

        Args:
            connection: psycopg.Connection
                The database connection to get the baseline version.

        Returns:
            packaging.version.Version | None: The baseline version.

        Raises:
            PumSchemaMigrationError: If the migration table does not exist or if no baseline version is found
            PumSchemaMigrationNoBaselineError: If the migration table does not exist
        """

        if not self.exists(connection=connection):
            raise PumSchemaMigrationError(
                f"{self.migration_table_identifier_str} table does not exist."
            )

        query = psycopg.sql.SQL(
            """
            SELECT version
            FROM {table}
            WHERE id = (
                SELECT id
                FROM {table}
                ORDER BY version DESC, date_installed DESC
                LIMIT 1
            )
        """
        )

        parameters = {
            "table": self.migration_table_identifier,
        }

        cursor = SqlContent(query).execute(connection, parameters=parameters)
        row = cursor._pum_results[0] if cursor._pum_results else None
        if row is None:
            raise PumSchemaMigrationNoBaselineError(
                f"Baseline version not found in the {self.migration_table_identifier_str} table."
            )
        return packaging.version.parse(row[0])

    def migration_details(self, connection: psycopg.Connection, version: str | None = None) -> dict:
        """Return the migration details from the migration table.

        Args:
            connection:
                The database connection to get the migration details.
            version:
                The version of the migration to get details for.
                If None, last migration is returned.

        Returns:
            dict: The migration details.

        Raises:
            PumSchemaMigrationError: If the migration table does not exist or if no migration details are found.
        """
        query = None
        if version is None:
            query = psycopg.sql.SQL(
                """
                SELECT *
                FROM {table}
                WHERE id = (
                        SELECT id
                        FROM {table}
                        ORDER BY version DESC, date_installed DESC
                        LIMIT 1
                    )
                ORDER BY date_installed DESC
            """
            )

            parameters = {
                "table": self.migration_table_identifier,
            }
        else:
            query = psycopg.sql.SQL(
                """
                SELECT *
                FROM {table}
                WHERE version = {version}
            """
            )

            parameters = {
                "table": self.migration_table_identifier,
                "version": psycopg.sql.Literal(version),
            }

        cursor = SqlContent(query).execute(connection, parameters=parameters)
        row = cursor._pum_results[0] if cursor._pum_results else None
        if row is None:
            raise PumSchemaMigrationError(
                f"Migration details not found for version {version} in the {self.migration_table_identifier_str} table."
            )
        return dict(zip([desc[0] for desc in cursor._pum_description], row, strict=False))

    def compare(self, connection: psycopg.Connection) -> int:
        """Compare the migrations details in the database to the changelogs in the source.

        Args:
            connection: The database connection to get the baseline version.
        Returns:
            int: -1 if database is behind, 0 if up to date.

        Raises:
            PumSchemaMigrationError: If there is a mismatch between the database and the source.
        """

        current_version = self.baseline(connection=connection)
        migration_details = self.migration_details(connection=connection)
        changelogs = [str(changelog.version) for changelog in self.config.changelogs()]

        # Check if the current migration version is in the changelogs
        if migration_details["version"] not in changelogs:
            raise PumSchemaMigrationError(
                f"Changelog for version {migration_details['version']} not found in the source."
            )

        # Check if there are newer changelogs than current version
        for changelog_version in changelogs:
            if packaging.version.parse(changelog_version) > current_version:
                return -1  # database is behind

        return 0  # database is up to date
